## Unsorted utility functions.
#' @include DataClasses.R

############################################################
## valueCount2ScanIndex
##
## @description Simple helper function that converts the number of values
## per scan/spectrum to an integer vector that can be passed to the base
## xcms functions/downstream C functions.
##
## @title Create index vector for internal C calls
## @param valCount Numeric vector representing the number of values per
## spectrum.
## @return An integer vector with the index (0-based) in the mz or intensity
## vectors indicating the start of a spectrum.
## @author Johannes Rainer
valueCount2ScanIndex <- function(valCount){
    ## Convert into 0 based.
    valCount <- cumsum(valCount)
    return(as.integer(c(0, valCount[-length(valCount)])))
}

############################################################
## useOriginalCode
##
## Simple function allowing the user to enable using the orignal
## code instead of the new implementations.
## This sets options.
##
##' @title Enable usage of old xcms code
##'
##' @description This function allows to enable the usage of old, partially
##' deprecated code from xcms by setting a corresponding global option. See
##' details for functions affected.
##'
##' @note Usage of old code is strongly dicouraged. This function is thought
##' to be used mainly in the transition phase from xcms to xcms version 3.
##' @details The functions/methods that will be affected by this are:
##' \itemize{
##' \item \code{\link{do_findChromPeaks_matchedFilter}}
##' }
##' @param x logical(1) to specify whether or not original
##' old code should be used in corresponding functions. If not provided the
##' function simply returns the value of the global option.
##' @return logical(1) indicating whether old code is being
##' used.
##' @author Johannes Rainer
useOriginalCode <- function(x) {
    if (missing(x)) {
        res <- options()$BioC$xcms$useOriginalCode
        if (is.null(res))
            return(FALSE)
        return(res)
    }
    if (!is.logical(x))
        stop("'x' has to be logical.")
    b_opts <- getOption("BioC")
    x_opts <- b_opts$xcms
    x_opts$useOriginalCode <- x[1]
    b_opts$xcms <-  x_opts
    options(BioC = b_opts)
    return(options()$BioC$xcms$useOriginalCode)
}

## .getOriginalFunction <- function(x) {
##     if (!any(names(.ORIGINAL_FUNCTIONS)))
## }
## .ORIGINAL_FUNCTIONS <- c(
##     matchedFilter = ".matchedFilter_orig"
## )

##' @title Copy the content from an environment to another one
##' This function copies the content of an environment into another one.
##' @param env environment from which to copy.
##' @param inheritLocks logical(1) whether the locking status should be copied
##' too
##' @return an env.
##' @noRd
.copy_env <- function(env, inheritLocks = FALSE) {
    new_e <- new.env(parent = emptyenv())
    eNames <- ls(env, all.names = TRUE)
    if (length(eNames) > 0) {
        for (eN in eNames) {
            new_e[[eN]] <- env[[eN]]
        }
    }
    if (inheritLocks) {
        if (environmentIsLocked(env))
            lockEnvironment(new_e)
    }
    return(new_e)
}

## #' Simulates the \code{findRange} function.
## #' @noRd
## findRangeR <- function(x, values) {
##     start <- min(which(x >= values[1]))
##     end <- max(which(x <= values[2]))
##     return(c(start, end))
## }

############################################################
## .createProfileMatrix
##' @title Create the profile matrix
##'
##' @description This function creates a \emph{profile} matrix, i.e. a rt times
##'     m/z matrix of aggregated intensity values with values aggregated within
##'     bins along the m/z dimension.
##'
##' @details This is somewhat the successor function for the deprecated
##'     \code{profBin} methods (\code{profBinM}, \code{profBinLinM},
##'     \code{profBinLinBaseM} and \code{profIntLin}).
##'
##' @param mz Numeric representing the m/z values across all scans/spectra.
##'
##' @param int Numeric representing the intensity values across all
##'     scans/spectra.
##'
##' @param valsPerSpect Numeric representing the number of measurements for each
##'     scan/spectrum.
##'
##' @param method A character string specifying the profile matrix generation
##'     method. Allowed are \code{"bin"}, \code{"binlin"},
##'     \code{"binlinbase"} and \code{"intlin"}.
##'
##' @param step Numeric specifying the size of the m/z bins.
##'
##' @param baselevel Numeric specifying the base value.
##'
##' @param basespace Numeric.
##'
##' @param mzrange. numeric(2) optionally specifying the mz value range
##'     for binning. This is to adopt the old profStepPad<- method used for
##'     obiwarp retention time correction that did the binning from
##'     whole-number limits.
##'
##' @param returnBreaks logical(1): hack to return the breaks of the bins.
##'     Setting this to TRUE causes the function to return a \code{list} with
##'     elements \code{"$profMat"} and \code{"breaks"}.
##'
##' @param baseValue numeric(1) defining the value to be returned if no signal
##'     was found in the corresponding bin. Defaults to 0 for backward
##'     compatibility.
##'
##' @noRd
.createProfileMatrix <- function(mz, int, valsPerSpect,
                                 method, step = 0.1, baselevel = NULL,
                                 basespace = NULL,
                                 mzrange. = NULL,
                                 returnBreaks = FALSE,
                                 baseValue = 0) {
    profMeths <- c("bin", "binlin", "binlinbase", "intlin")
    names(profMeths) <- c("none", "lin", "linbase", "intlin")
    method <- match.arg(method, profMeths)
    impute <- names(profMeths)[profMeths == method]
    brks <- NULL
    
    if (length(mzrange.) != 2) {
        mrange <- range(mz, na.rm = TRUE)
        mzrange. <- c(floor(mrange[1] / step) * step,
                      ceiling(mrange[2] / step) * step)
    }
    mass <- seq(mzrange.[1], mzrange.[2], by = step)
    mlength <- length(mass)
    ## Calculate the "real" bin size; old xcms code oddity that that's different
    ## from step.
    bin_size <- (mass[mlength] - mass[1]) / (mlength - 1)
    ## Define the breaks.
    toIdx <- cumsum(valsPerSpect)
    fromIdx <- c(1L, toIdx[-length(toIdx)] + 1L)
    shiftBy <- TRUE
    binFromX <- min(mass)
    binToX <- max(mass)
    brks <- breaks_on_nBins(fromX = binFromX, toX = binToX,
                            nBins = mlength, shiftByHalfBinSize = TRUE)
    ## for profIntLinM we have to use the old code.
    if (impute == "intlin") {
        profFun <- "profIntLinM"
        profp <- list()
        scanindex <- valueCount2ScanIndex(valsPerSpect)
        buf <- do.call(profFun, args = list(mz, int,
                                            scanindex, mlength,
                                            mass[1], mass[mlength],
                                            TRUE))
    } else {
        ## Binning the data.
        binRes <- binYonX(mz, int,
                          breaks = brks,
                          fromIdx = fromIdx,
                          toIdx = toIdx,
                          baseValue = ifelse(impute == "none", yes = baseValue,
                                             no = NA),
                          sortedX = TRUE,
                          returnIndex = FALSE
                          )
        if (length(toIdx) == 1)
            binRes <- list(binRes)
        ## Missing value imputation.
        if (impute == "linbase") {
            ## need arguments distance and baseValue.
            if (length(basespace) > 0) {
                if (!is.numeric(basespace))
                    stop("'basespace' has to be numeric!")
                distance <- floor(basespace[1] / bin_size)
            } else {
                distance <- floor(0.075 / bin_size)
            }
            if (length(baselevel) > 0) {
                if (!is.numeric(baselevel))
                    stop("'baselevel' has to be numeric!")
                baseValue <- baselevel
            } else {
                baseValue <- min(int, na.rm = TRUE) / 2
            }
        } else {
            distance <- 0
            baseValue <- 0
        }
        binVals <- lapply(binRes, function(z) {
            return(imputeLinInterpol(z$y, method = impute, distance = distance,
                                     noInterpolAtEnds = TRUE,
                                     baseValue = baseValue))
        })
        buf <- do.call(cbind, binVals)
    }
    if (returnBreaks)
        buf <- list(profMat = buf, breaks = brks)
    buf
}

#' @description This function creates arbitrary IDs for features.
#' 
#' @param x integer(1) with the number of IDs that should be generated.
#'
#' @noRd
.featureIDs <- function(x) {
    sprintf(paste0("FT%0", ceiling(log10(x + 1L)), "d"), 1:x)
}
